@@ -25,13 +25,15 @@ import org.apache.commons.pool.TestBaseKeyedObjectPool;
 import org.apache.commons.pool.VisitTracker;
 import org.apache.commons.pool.VisitTrackerFactory;
 
+import java.io.PrintWriter;
+import java.io.StringWriter;
 import java.util.HashMap;
 import java.util.NoSuchElementException;
 import java.util.Random;
 
 /**
  * @author Rodney Waldhoff
- * @version $Revision: 606718 $ $Date: 2007-12-24 10:55:10 -0700 (Mon, 24 Dec 2007) $
+ * @version $Revision: 791860 $ $Date: 2009-07-07 11:10:30 -0400 (Tue, 07 Jul 2009) $
  */
 public class TestGenericKeyedObjectPool extends TestBaseKeyedObjectPool {
     public TestGenericKeyedObjectPool(String testName) {
@@ -79,9 +81,9 @@ public class TestGenericKeyedObjectPool extends TestBaseKeyedObjectPool {
     }
 
     private GenericKeyedObjectPool pool = null;
-    private Integer zero = new Integer(0);
-    private Integer one = new Integer(1);
-    private Integer two = new Integer(2);
+    private final Integer zero = new Integer(0);
+    private final Integer one = new Integer(1);
+    private final Integer two = new Integer(2);
 
     public void setUp() throws Exception {
         super.setUp();
@@ -90,6 +92,7 @@ public class TestGenericKeyedObjectPool extends TestBaseKeyedObjectPool {
 
     public void tearDown() throws Exception {
         super.tearDown();
+        pool.clear();
         pool.close();
         pool = null;
     }
@@ -211,6 +214,15 @@ public class TestGenericKeyedObjectPool extends TestBaseKeyedObjectPool {
             // expected
         }
     }
+    
+    public void testWhenExhaustedGrow() throws Exception {
+        pool.setMaxActive(1);
+        pool.setMaxTotal(1);
+        pool.setWhenExhaustedAction(GenericObjectPool.WHEN_EXHAUSTED_GROW);
+        for (int i = 0; i < 10; i++) {
+            pool.borrowObject("a");
+        }
+    }
 
     public void testMaxTotal() throws Exception {
         pool.setMaxActive(2);
@@ -386,17 +398,17 @@ public class TestGenericKeyedObjectPool extends TestBaseKeyedObjectPool {
             pool.returnObject("",active[i]);
         }
 
-        try { Thread.sleep(1000L); } catch(Exception e) { }
+        try { Thread.sleep(1000L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 500 idle, found " + pool.getNumIdle(""),pool.getNumIdle("") < 500);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 400 idle, found " + pool.getNumIdle(""),pool.getNumIdle("") < 400);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 300 idle, found " + pool.getNumIdle(""),pool.getNumIdle("") < 300);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 200 idle, found " + pool.getNumIdle(""),pool.getNumIdle("") < 200);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 100 idle, found " + pool.getNumIdle(""),pool.getNumIdle("") < 100);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertEquals("Should be zero idle, found " + pool.getNumIdle(""),0,pool.getNumIdle(""));
 
         for(int i=0;i<500;i++) {
@@ -406,17 +418,17 @@ public class TestGenericKeyedObjectPool extends TestBaseKeyedObjectPool {
             pool.returnObject("",active[i]);
         }
 
-        try { Thread.sleep(1000L); } catch(Exception e) { }
+        try { Thread.sleep(1000L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 500 idle, found " + pool.getNumIdle(""),pool.getNumIdle("") < 500);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 400 idle, found " + pool.getNumIdle(""),pool.getNumIdle("") < 400);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 300 idle, found " + pool.getNumIdle(""),pool.getNumIdle("") < 300);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 200 idle, found " + pool.getNumIdle(""),pool.getNumIdle("") < 200);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 100 idle, found " + pool.getNumIdle(""),pool.getNumIdle("") < 100);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertEquals("Should be zero idle, found " + pool.getNumIdle(""),0,pool.getNumIdle(""));
     }
 
@@ -438,27 +450,27 @@ public class TestGenericKeyedObjectPool extends TestBaseKeyedObjectPool {
             pool.returnObject("2",active2[i]);
         }
 
-        try { Thread.sleep(1000L); } catch(Exception e) { }
+        try { Thread.sleep(1100L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 1000 idle, found " + pool.getNumIdle(),pool.getNumIdle() < 1000);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 900 idle, found " + pool.getNumIdle(),pool.getNumIdle() < 900);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 800 idle, found " + pool.getNumIdle(),pool.getNumIdle() < 800);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 700 idle, found " + pool.getNumIdle(),pool.getNumIdle() < 700);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 600 idle, found " + pool.getNumIdle(),pool.getNumIdle() < 600);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 500 idle, found " + pool.getNumIdle(),pool.getNumIdle() < 500);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 400 idle, found " + pool.getNumIdle(),pool.getNumIdle() < 400);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 300 idle, found " + pool.getNumIdle(),pool.getNumIdle() < 300);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 200 idle, found " + pool.getNumIdle(),pool.getNumIdle() < 200);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertTrue("Should be less than 100 idle, found " + pool.getNumIdle(),pool.getNumIdle() < 100);
-        try { Thread.sleep(600L); } catch(Exception e) { }
+        try { Thread.sleep(600L); } catch(InterruptedException e) { }
         assertEquals("Should be zero idle, found " + pool.getNumIdle(),0,pool.getNumIdle());
     }
 
@@ -478,12 +490,12 @@ public class TestGenericKeyedObjectPool extends TestBaseKeyedObjectPool {
             while(!(threads[i]).complete()) {
                 try {
                     Thread.sleep(500L);
-                } catch(Exception e) {
+                } catch(InterruptedException e) {
                     // ignored
                 }
             }
             if(threads[i].failed()) {
-                fail();
+                fail("Thread failed: "+i+"\n"+getExceptionTrace(threads[i]._exception));
             }
         }
     }
@@ -530,27 +542,27 @@ public class TestGenericKeyedObjectPool extends TestBaseKeyedObjectPool {
 
         pool.preparePool(key, true);
 
-        try { Thread.sleep(150L); } catch(Exception e) { }
+        try { Thread.sleep(150L); } catch(InterruptedException e) { }
         assertTrue("Should be 5 idle, found " + pool.getNumIdle(),pool.getNumIdle() == 5);
 
         Object[] active = new Object[5];
         active[0] = pool.borrowObject(key);
 
-        try { Thread.sleep(150L); } catch(Exception e) { }
+        try { Thread.sleep(150L); } catch(InterruptedException e) { }
         assertTrue("Should be 5 idle, found " + pool.getNumIdle(),pool.getNumIdle() == 5);
 
         for(int i=1 ; i<5 ; i++) {
             active[i] = pool.borrowObject(key);
         }
 
-        try { Thread.sleep(150L); } catch(Exception e) { }
+        try { Thread.sleep(150L); } catch(InterruptedException e) { }
         assertTrue("Should be 5 idle, found " + pool.getNumIdle(),pool.getNumIdle() == 5);
 
         for(int i=0 ; i<5 ; i++) {
             pool.returnObject(key, active[i]);
         }
 
-        try { Thread.sleep(150L); } catch(Exception e) { }
+        try { Thread.sleep(150L); } catch(InterruptedException e) { }
         assertTrue("Should be 10 idle, found " + pool.getNumIdle(),pool.getNumIdle() == 10);
     }
 
@@ -569,40 +581,40 @@ public class TestGenericKeyedObjectPool extends TestBaseKeyedObjectPool {
         assertTrue("Should be 5 idle, found " + 
                 pool.getNumIdle(),pool.getNumIdle() == 5);
 
-        try { Thread.sleep(150L); } catch(Exception e) { }
+        try { Thread.sleep(150L); } catch(InterruptedException e) { }
         assertTrue("Should be 5 idle, found " + pool.getNumIdle(),pool.getNumIdle() == 5);
 
         Object[] active = new Object[10];
 
-        try { Thread.sleep(150L); } catch(Exception e) { }
+        try { Thread.sleep(150L); } catch(InterruptedException e) { }
         assertTrue("Should be 5 idle, found " + pool.getNumIdle(),pool.getNumIdle() == 5);
 
         for(int i=0 ; i<5 ; i++) {
             active[i] = pool.borrowObject(key);
         }
 
-        try { Thread.sleep(150L); } catch(Exception e) { }
+        try { Thread.sleep(150L); } catch(InterruptedException e) { }
         assertTrue("Should be 5 idle, found " + pool.getNumIdle(),pool.getNumIdle() == 5);
 
         for(int i=0 ; i<5 ; i++) {
             pool.returnObject(key, active[i]);
         }
 
-        try { Thread.sleep(150L); } catch(Exception e) { }
+        try { Thread.sleep(150L); } catch(InterruptedException e) { }
         assertTrue("Should be 10 idle, found " + pool.getNumIdle(),pool.getNumIdle() == 10);
 
         for(int i=0 ; i<10 ; i++) {
             active[i] = pool.borrowObject(key);
         }
 
-        try { Thread.sleep(150L); } catch(Exception e) { }
+        try { Thread.sleep(150L); } catch(InterruptedException e) { }
         assertTrue("Should be 0 idle, found " + pool.getNumIdle(),pool.getNumIdle() == 0);
 
         for(int i=0 ; i<10 ; i++) {
             pool.returnObject(key, active[i]);
         }
 
-        try { Thread.sleep(150L); } catch(Exception e) { }
+        try { Thread.sleep(150L); } catch(InterruptedException e) { }
         assertTrue("Should be 10 idle, found " + pool.getNumIdle(),pool.getNumIdle() == 10);
     }
 
@@ -623,7 +635,7 @@ public class TestGenericKeyedObjectPool extends TestBaseKeyedObjectPool {
 
         assertTrue("Should be 0 idle, found " + pool.getNumIdle(),pool.getNumIdle() == 0);
 
-        try { Thread.sleep(1500L); } catch(Exception e) { }
+        try { Thread.sleep(1500L); } catch(InterruptedException e) { }
         assertTrue("Should be 5 idle, found " + pool.getNumIdle(),pool.getNumIdle() == 5);
     }
 
@@ -640,13 +652,13 @@ public class TestGenericKeyedObjectPool extends TestBaseKeyedObjectPool {
         //Generate a random key
         String key = "A";
 
-        try { Thread.sleep(150L); } catch(Exception e) { }
+        try { Thread.sleep(150L); } catch(InterruptedException e) { }
         assertTrue("Should be 0 idle, found " + pool.getNumIdle(),pool.getNumIdle() == 0);
 
         Object active = pool.borrowObject(key);
         assertNotNull(active);
 
-        try { Thread.sleep(150L); } catch(Exception e) { }
+        try { Thread.sleep(150L); } catch(InterruptedException e) { }
         assertTrue("Should be 5 idle, found " + pool.getNumIdle(),pool.getNumIdle() == 5);
     }
 
@@ -886,10 +898,11 @@ public class TestGenericKeyedObjectPool extends TestBaseKeyedObjectPool {
         Random random = new Random();
         random.setSeed(System.currentTimeMillis());
         pool.setMaxIdle(-1);
-        for (int i = 0; i < 4; i++) {
+        for (int i = 0; i < smallPrimes.length; i++) {
             pool.setNumTestsPerEvictionRun(smallPrimes[i]);
-            for (int j = 0; j < 5; j++) {
+            for (int j = 0; j < 5; j++) {// Try the tests a few times
                 pool.clear();
+                assertEquals("NumIdle should be zero after clearing the pool",0,pool.getNumIdle());
                 int zeroLength = 10 + random.nextInt(20);
                 for (int k = 0; k < zeroLength; k++) {
                     pool.addObject(zero);
@@ -922,22 +935,37 @@ public class TestGenericKeyedObjectPool extends TestBaseKeyedObjectPool {
                 int visitCount = 0;
                 for (int k = 0; k < zeroLength; k++) {
                     tracker = (VisitTracker) pool.borrowObject(zero); 
-                    visitCount = tracker.getValidateCount();                  
-                    assertTrue(visitCount >= cycleCount && 
-                            visitCount <= cycleCount + 1);
+                    visitCount = tracker.getValidateCount();
+                    if (visitCount < cycleCount || visitCount > cycleCount + 1){
+                        fail(formatSettings("ZERO", "runs", runs, "lifo", lifo, "i", i, "j", j,
+                                "k", k, "visitCount", visitCount, "cycleCount", cycleCount,
+                                "totalInstances", totalInstances, zeroLength, oneLength, twoLength));
+                    }
                 }
                 for (int k = 0; k < oneLength; k++) {
                     tracker = (VisitTracker) pool.borrowObject(one); 
                     visitCount = tracker.getValidateCount();
-                    assertTrue(visitCount >= cycleCount && 
-                            visitCount <= cycleCount + 1);
+                    if (visitCount < cycleCount || visitCount > cycleCount + 1){
+                        fail(formatSettings("ONE", "runs", runs, "lifo", lifo, "i", i, "j", j,
+                                "k", k, "visitCount", visitCount, "cycleCount", cycleCount,
+                                "totalInstances", totalInstances, zeroLength, oneLength, twoLength));
+                    }
                 }
+                int visits[] = new int[twoLength];
                 for (int k = 0; k < twoLength; k++) {
                     tracker = (VisitTracker) pool.borrowObject(two); 
                     visitCount = tracker.getValidateCount();
-                    assertTrue(visitCount >= cycleCount && 
-                            visitCount <= cycleCount + 1);
-                } 
+                    visits[k] = visitCount;
+                    if (visitCount < cycleCount || visitCount > cycleCount + 1){
+                        StringBuffer sb = new StringBuffer("Visits:");
+                        for (int l = 0; l <= k; l++){
+                            sb.append(visits[l]).append(' ');
+                        }
+                        fail(formatSettings("TWO "+sb.toString(), "runs", runs, "lifo", lifo, "i", i, "j", j,
+                                "k", k, "visitCount", visitCount, "cycleCount", cycleCount,
+                                "totalInstances", totalInstances, zeroLength, oneLength, twoLength));
+                    }
+                }
             }
         }
     }
@@ -1205,22 +1233,153 @@ public class TestGenericKeyedObjectPool extends TestBaseKeyedObjectPool {
         assertEquals(whenExhaustedAction,pool.getWhenExhaustedAction());
         assertEquals(lifo, pool.getLifo());  
     }
+    
+    public void testExceptionOnPassivateDuringReturn() throws Exception {
+        SimpleFactory factory = new SimpleFactory();        
+        GenericKeyedObjectPool pool = new GenericKeyedObjectPool(factory);
+        Object obj = pool.borrowObject("one");
+        factory.setThrowExceptionOnPassivate(true);
+        pool.returnObject("one", obj);
+        assertEquals(0,pool.getNumIdle());
+        pool.close();
+    }
+    
+    public void testExceptionOnDestroyDuringBorrow() throws Exception {
+        SimpleFactory factory = new SimpleFactory(); 
+        factory.setThrowExceptionOnDestroy(true);
+        factory.setValidationEnabled(true);
+        GenericKeyedObjectPool pool = new GenericKeyedObjectPool(factory);
+        pool.setTestOnBorrow(true);
+        pool.borrowObject("one");
+        factory.setValid(false); // Make validation fail on next borrow attempt
+        try {
+            pool.borrowObject("one");
+            fail("Expecting NoSuchElementException");
+        } catch (NoSuchElementException ex) {
+            // expected
+        }
+        assertEquals(1, pool.getNumActive("one"));
+        assertEquals(0, pool.getNumIdle("one"));
+        assertEquals(1, pool.getNumActive());
+        assertEquals(0, pool.getNumIdle());
+    }
+    
+    public void testExceptionOnDestroyDuringReturn() throws Exception {
+        SimpleFactory factory = new SimpleFactory(); 
+        factory.setThrowExceptionOnDestroy(true);
+        factory.setValidationEnabled(true);
+        GenericKeyedObjectPool pool = new GenericKeyedObjectPool(factory);
+        pool.setTestOnReturn(true);
+        Object obj1 = pool.borrowObject("one");
+        pool.borrowObject("one");
+        factory.setValid(false); // Make validation fail
+        pool.returnObject("one", obj1);
+        assertEquals(1, pool.getNumActive("one"));
+        assertEquals(0, pool.getNumIdle("one"));
+        assertEquals(1, pool.getNumActive());
+        assertEquals(0, pool.getNumIdle());
+    }
+    
+    public void testExceptionOnActivateDuringBorrow() throws Exception {
+        SimpleFactory factory = new SimpleFactory(); 
+        GenericKeyedObjectPool pool = new GenericKeyedObjectPool(factory);
+        Object obj1 = pool.borrowObject("one");
+        Object obj2 = pool.borrowObject("one");
+        pool.returnObject("one", obj1);
+        pool.returnObject("one", obj2);
+        factory.setThrowExceptionOnActivate(true);
+        factory.setEvenValid(false);  
+        // Activation will now throw every other time
+        // First attempt throws, but loop continues and second succeeds
+        Object obj = pool.borrowObject("one");
+        assertEquals(1, pool.getNumActive("one"));
+        assertEquals(0, pool.getNumIdle("one"));
+        assertEquals(1, pool.getNumActive());
+        assertEquals(0, pool.getNumIdle());
+        
+        pool.returnObject("one", obj);
+        factory.setValid(false);
+        // Validation will now fail on activation when borrowObject returns
+        // an idle instance, and then when attempting to create a new instance
+        try {
+            pool.borrowObject("one");
+            fail("Expecting NoSuchElementException");
+        } catch (NoSuchElementException ex) {
+            // expected
+        }
+        assertEquals(0, pool.getNumActive("one"));
+        assertEquals(0, pool.getNumIdle("one"));
+        assertEquals(0, pool.getNumActive());
+        assertEquals(0, pool.getNumIdle());
+    }
+    
+    public void testBlockedKeyDoesNotBlockPool() throws Exception {
+        SimpleFactory factory = new SimpleFactory();
+        GenericKeyedObjectPool pool = new GenericKeyedObjectPool(factory);
+        pool.setWhenExhaustedAction(GenericObjectPool.WHEN_EXHAUSTED_BLOCK);
+        pool.setMaxWait(5000);
+        pool.setMaxActive(1);
+        pool.setMaxTotal(-1);
+        pool.borrowObject("one");
+        long start = System.currentTimeMillis();
+        // Needs to be in a separate thread as this will block
+        Runnable simple = new SimpleTestThread(pool, "one");
+        (new Thread(simple)).start();
+        // This should be almost instant. If it isn't it means this thread got
+        // stuck behind the thread created above which is bad.
+        // Give other thread a chance to start
+        Thread.sleep(1000);
+        pool.borrowObject("two");
+        long end = System.currentTimeMillis();
+        // If it fails it will be more than 4000ms (5000 less the 1000 sleep)
+        // If it passes it should be almost instant
+        // Use 3000ms as the threshold - should avoid timing issues on most
+        // (all? platforms)
+        assertTrue ((end-start) < 4000);
+        
+    }
 
-    class TestThread implements Runnable {
-        java.util.Random _random = new java.util.Random();
-        KeyedObjectPool _pool = null;
-        boolean _complete = false;
-        boolean _failed = false;
-        int _iter = 100;
-        int _delay = 50;
+    /*
+     * Very simple test thread that just tries to borrow an object from
+     * the provided pool with the specified key and returns it
+     */
+    static class SimpleTestThread implements Runnable {
+        private final KeyedObjectPool _pool;
+        private final String _key;
+        
+        public SimpleTestThread(KeyedObjectPool pool, String key) {
+            _pool = pool;
+            _key = key;
+        }
+
+        public void run() {
+            try {
+                Object obj = _pool.borrowObject(_key);
+                _pool.returnObject(_key, obj);
+            } catch (Exception e) {
+                // Ignore
+            }
+        }
+    }
+    
+    static class TestThread implements Runnable {
+        private final java.util.Random _random = new java.util.Random();
+        
+        // Thread config items
+        private final KeyedObjectPool _pool;
+        private final int _iter;
+        private final int _delay;
+
+        private volatile boolean _complete = false;
+        private volatile boolean _failed = false;
+        private volatile Exception _exception;
 
         public TestThread(KeyedObjectPool pool) {
-            _pool = pool;
+            this(pool, 100, 50);
         }
 
         public TestThread(KeyedObjectPool pool, int iter) {
-            _pool = pool;
-            _iter = iter;
+            this(pool, iter, 50);
         }
 
         public TestThread(KeyedObjectPool pool, int iter, int delay) {
@@ -1241,29 +1400,29 @@ public class TestGenericKeyedObjectPool extends TestBaseKeyedObjectPool {
             for(int i=0;i<_iter;i++) {
                 String key = String.valueOf(_random.nextInt(3));
                 try {
-                    Thread.sleep((long)_random.nextInt(_delay));
-                } catch(Exception e) {
+                    Thread.sleep(_random.nextInt(_delay));
+                } catch(InterruptedException e) {
                     // ignored
                 }
                 Object obj = null;
                 try {
                     obj = _pool.borrowObject(key);
                 } catch(Exception e) {
-                    e.printStackTrace();
+                    _exception = e;
                     _failed = true;
                     _complete = true;
                     break;
                 }
 
                 try {
-                    Thread.sleep((long)_random.nextInt(_delay));
-                } catch(Exception e) {
+                    Thread.sleep(_random.nextInt(_delay));
+                } catch(InterruptedException e) {
                     // ignored
                 }
                 try {
                     _pool.returnObject(key,obj);
                 } catch(Exception e) {
-                    e.printStackTrace();
+                    _exception = e;
                     _failed = true;
                     _complete = true;
                     break;
@@ -1290,11 +1449,14 @@ public class TestGenericKeyedObjectPool extends TestBaseKeyedObjectPool {
             }
             return String.valueOf(key) + String.valueOf(counter++);
         }
-        public void destroyObject(Object key, Object obj) {
+        public void destroyObject(Object key, Object obj) throws Exception {
             doWait(destroyLatency);
             synchronized(this) {
                 activeCount--;
             }
+            if (exceptionOnDestroy) {
+                throw new Exception();
+            }
         }
         public boolean validateObject(Object key, Object obj) {
             if (enableValidation) { 
@@ -1303,8 +1465,18 @@ public class TestGenericKeyedObjectPool extends TestBaseKeyedObjectPool {
                 return valid;
             }
         }
-        public void activateObject(Object key, Object obj) { }
-        public void passivateObject(Object key, Object obj) { }
+        public void activateObject(Object key, Object obj) throws Exception {
+            if (exceptionOnActivate) {
+                if (!(validateCounter++%2 == 0 ? evenValid : oddValid)) {
+                    throw new Exception();
+                }
+            }
+        }
+        public void passivateObject(Object key, Object obj) throws Exception {
+            if (exceptionOnPassivate) {
+                throw new Exception();
+            }
+        }
         
         public void setMaxActive(int maxActive) {
             this.maxActive = maxActive;
@@ -1318,6 +1490,22 @@ public class TestGenericKeyedObjectPool extends TestBaseKeyedObjectPool {
         void setEvenValid(boolean valid) {
             evenValid = valid;
         }
+        void setValid(boolean valid) {
+            evenValid = valid;
+            oddValid = valid;
+        }
+        
+        public void setThrowExceptionOnActivate(boolean b) {
+            exceptionOnActivate = b;
+        }
+        
+        public void setThrowExceptionOnDestroy(boolean b) {
+            exceptionOnDestroy = b;
+        }
+        
+        public void setThrowExceptionOnPassivate(boolean b) {
+            exceptionOnPassivate = b;
+        }
         
         int counter = 0;
         boolean valid;
@@ -1329,6 +1517,9 @@ public class TestGenericKeyedObjectPool extends TestBaseKeyedObjectPool {
         boolean enableValidation = false;
         long destroyLatency = 0;
         int maxActive = Integer.MAX_VALUE;
+        boolean exceptionOnPassivate = false;
+        boolean exceptionOnActivate = false;
+        boolean exceptionOnDestroy = false;
         
         private void doWait(long latency) {
             try {
@@ -1347,6 +1538,28 @@ public class TestGenericKeyedObjectPool extends TestBaseKeyedObjectPool {
         return false;
     }
 
+    private String getExceptionTrace(Throwable t){
+        StringWriter sw = new StringWriter();
+        t.printStackTrace(new PrintWriter(sw));
+        return sw.toString();
+    }
+    
+    private String formatSettings(String title, String s, int i, String s0, boolean b0, String s1, int i1, String s2, int i2, String s3, int i3,
+            String s4, int i4, String s5, int i5, String s6, int i6, int zeroLength, int oneLength, int twoLength){
+        StringBuffer sb = new StringBuffer(80);
+        sb.append(title).append(' ');
+        sb.append(s).append('=').append(i).append(' ');
+        sb.append(s0).append('=').append(b0).append(' ');
+        sb.append(s1).append('=').append(i1).append(' ');
+        sb.append(s2).append('=').append(i2).append(' ');
+        sb.append(s3).append('=').append(i3).append(' ');
+        sb.append(s4).append('=').append(i4).append(' ');
+        sb.append(s5).append('=').append(i5).append(' ');
+        sb.append(s6).append('=').append(i6).append(' ');
+        sb.append("Lengths=").append(zeroLength).append(',').append(oneLength).append(',').append(twoLength).append(' ');
+        return sb.toString();
+    }
+    
 }
 
 
