@@ -34,7 +34,7 @@ import java.util.Random;
  * @author Rodney Waldhoff
  * @author Dirk Verbeeck
  * @author Sandy McArthur
- * @version $Revision: 791860 $ $Date: 2009-07-07 11:10:30 -0400 (Tue, 07 Jul 2009) $
+ * @version $Revision: 831698 $ $Date: 2009-11-01 11:33:31 -0500 (Sun, 01 Nov 2009) $
  */
 public class TestGenericObjectPool extends TestBaseObjectPool {
     public TestGenericObjectPool(String testName) {
@@ -1458,4 +1458,51 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
             }
         }
     }
+    
+    /**
+     * On first borrow, first object fails validation, second object is OK.
+     * Subsequent borrows are OK. This was POOL-152.
+     */
+    public void testBrokenFactoryShouldNotBlockPool() {
+        int maxActive = 1;
+        
+        SimpleFactory factory = new SimpleFactory();
+        factory.setMaxActive(maxActive);
+        pool.setFactory(factory);
+        pool.setMaxActive(maxActive);
+        pool.setWhenExhaustedAction(GenericObjectPool.WHEN_EXHAUSTED_BLOCK);
+        pool.setTestOnBorrow(true);
+        
+        // First borrow object will need to create a new object which will fail
+        // validation.
+        Object obj = null;
+        Exception ex = null;
+        factory.setValid(false);
+        try {
+            obj = pool.borrowObject();
+        } catch (Exception e) {
+            ex = e;
+        }
+        // Failure expected
+        assertNotNull(ex);
+        assertTrue(ex instanceof NoSuchElementException);
+        assertNull(obj);
+
+        // Configure factory to create valid objects so subsequent borrows work
+        factory.setValid(true);
+        
+        // Subsequent borrows should be OK
+        try {
+            obj = pool.borrowObject();
+        } catch (Exception e1) {
+            fail();
+        }
+        assertNotNull(obj);
+        try {
+            pool.returnObject(obj);
+        } catch (Exception e) {
+            fail();
+        }
+        
+    }
 }
