@@ -23,12 +23,11 @@ import org.apache.commons.pool.impl.GenericObjectPool;
  * Multi-thread performance test
  * 
  * @author Dirk Verbeeck
- * @version $Revision: 480413 $ $Date: 2006-11-28 22:16:05 -0700 (Tue, 28 Nov 2006) $ 
+ * @version $Revision: 775703 $ $Date: 2009-05-17 12:39:51 -0400 (Sun, 17 May 2009) $ 
  */
 public class PerformanceTest {
     private int logLevel = 0;
     private int nrIterations = 5;
-    private int nrThreads = 100;
 
     private GenericObjectPool pool;
     private boolean start = false;
@@ -108,7 +107,6 @@ public class PerformanceTest {
 
     private void run(int nrIterations, int nrThreads, int maxActive, int maxIdle) {
         this.nrIterations = nrIterations;
-        this.nrThreads = nrThreads;
         init();
         
         SleepingObjectFactory factory = new SleepingObjectFactory();
